#!/usr/bin/env python
"""#
# for local install:
#   fab -H root@localhost setup
#
# for test
#   env SAAS_USER=accounts fab -H root@accounts.test.openerp.com update:saas-1
#"""
from __future__ import print_function

# STOP THE WARNINGS!!
import warnings
with warnings.catch_warnings():
    import cryptography

import base64
from collections import defaultdict, OrderedDict, namedtuple
import datetime
from functools import wraps, partial
import hashlib
import io
import os
import passlib.utils
import re
import signal
import sys
import time
from uuid import uuid4
import werkzeug
import zlib

from fabric.api import task, env, run, cd, sudo, put, hide, hosts, local, get, execute
from fabric.colors import red, yellow, green, blue, white
from fabric.context_managers import warn_only, settings, shell_env
from fabric.contrib.console import confirm as _confirm
from fabric.contrib.files import exists, upload_template
from fabric.decorators import roles, runs_once, with_settings
from fabric.operations import prompt
from fabric.utils import abort, puts, fastprint, warn

import fabtools
import fabtools.require
from fabtools.require.files import temporary_directory

from dateutil.relativedelta import relativedelta
import json

#----------------------------------------------------------
# environment
#----------------------------------------------------------

USER = os.environ.get('SAAS_USER', 'odoo')
if not re.match(r'^[a-z_]+$', USER):
    abort('%r is not alphabetical' % (USER,))

HOME = '/home/%s' % USER

ODOO_SRC_PATH = HOME + '/src'
HERE = os.path.dirname(os.path.realpath(__file__))

env.use_ssh_config = True
env.sudo_user = USER
env.sudo_prefix += '-H '


#----------------------------------------------------------
# signal handling
#----------------------------------------------------------

_STOPPED = False
_STOPPABLE = False

def stopped():
    return _STOPPED

def graceful_stop(f):
    @wraps(f)
    def w(*a, **kw):
        global _STOPPABLE
        _STOPPABLE = True
        return f(*a, **kw)
    return w

def check_stop(it):
    for i in it:
        if _STOPPED:
            break
        yield i

def _sigint_handler(s, f):
    global _STOPPED
    if _STOPPABLE and not _STOPPED:
        puts('CTRL-C catched. Will Stop after next operation')
        _STOPPED = True
    else:
        if _STOPPABLE:
            puts('CTRL-C catched. Stop immediatly')
        raise KeyboardInterrupt()

signal.signal(signal.SIGINT, _sigint_handler)

#----------------------------------------------------------
# tools
#----------------------------------------------------------
def str2bool(s):
    s = s.lower()
    y = 'y yes 1 true t'.split()
    n = 'n no 0 false f'.split()
    assert s in y + n, 'Use yes/no/true/false'
    return s in y

def sizeof_fmt(num, suffix='B'):
    # http://stackoverflow.com/a/1094933/75349
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f %s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f %s%s" % (num, 'Yi', suffix)

def ensure_xz_file(file_, url, md5=None, force=False):
    if force:
        md5 = 'force-download'
    dir_, file_ = os.path.split(file_)
    with cd(dir_):
        xz = file_ + '.xz'

        def decompress():
            run('mv {0} {1}'.format(file_, xz))  # to allow decompression
            run('xz -d {0}'.format(xz))

        with fabtools.files.watch(file_, callback=decompress):
            fabtools.require.file(path=file_, url=url, md5=md5)

def as_(user):
    def deco(fun):
        @wraps(fun)
        def wrapper(*args, **kwargs):
            current_user = env.user
            if current_user != user:
                warn('force using user ' + user)
            env.user = user
            try:
                return fun(*args, **kwargs)
            finally:
                env.user = current_user
        return wrapper
    return deco

def hide_disconnect_messages(fn):
    @wraps(fn)
    def wrapper(*a, **kw):
        try:
            return fn(*a, **kw)
        finally:
            from fabric.state import output
            output.status = False
    return wrapper

def get_port(branch, with_sudo=True):
    func = sudo if with_sudo else run
    return int(func('{0}/bin/oe-meta get-port {1}'.format(HOME, branch)).strip())

def all_branches(all_=False, with_sudo=True):
    flags = '-a' if all_ else ''
    func = sudo if with_sudo else run
    return func('{0}/bin/oe-meta list-branches {1}'.format(HOME, flags)).strip().split()

def all_dbs(user, with_sudo=False):
    func = sudo if with_sudo else run
    with hide('stdout'):
        with cd('~'):
            out = func('psql -l -t -AF/ | grep "/{user}/" | grep -v "^meta/" | cut -d/ -f1'.format(user=user))
    return out.split()

def friday_check(action='make this action'):
    if datetime.date.today().weekday() != 4:
        return
    if hasattr(env, 'friday_check'):
        # already asked (and it's ok, else it would have been aboted)
        return
    answer = prompt(red("We are friday. Do you really want to {0}?".format(action)),
                    key='friday_check',
                    validate=str2bool)
    if not answer:
        abort(yellow('Have a nice week-end!'))

def has_systemd():
    with hide('output', 'running', 'warnings'):
        with settings(warn_only=True):
            res = run('command -v systemctl')
            return res.return_code == 0

def branch2service(branch):
    """returns a tuple (service_name, service_path)"""
    service_name = 'openerp-' + branch
    if has_systemd():
        service_path = '/etc/systemd/system/{0}.service'.format(service_name)
    else:
        service_path = '/etc/init.d/{0}'.format(service_name)
    return (service_name, service_path)

def psql_query(db, sql=None, cache=False, first=False, **_settings):
    """
    Requires Postgresql >= 9.2

    TODO fme: test behavior with insert, update, ... + add doc + submit to fabtools.fabric
    """
    sql = 'SELECT array_to_json(array_agg(query)) FROM (%s) AS query' % sql
    if cache:
        if 'tabat_hash_cache' not in env:
            env['tabat_hash_cache'] = {}
        # compute a hash containing all result dependent arguments
        fingerprint = db + sql + str(first) + _settings.get('host_string') or env.host_string
        tabat_hash_cache = hashlib.md5(fingerprint).hexdigest()
        if tabat_hash_cache in env['tabat_hash_cache']:
            return env['tabat_hash_cache'][tabat_hash_cache]

    cmd = 'psql -tAqX --pset=pager=off -d {0} -c "{1}"'.format(db, sql.replace('"', '\\"'))

    with settings(hide('stdout'), **_settings):
        res = run(cmd)
    if not res:
        res = None
    else:
        res = json.loads(res)

    if res is not None and first:
        if not len(res):
            res = None
        else:
            vals = res[0].values()
            if len(vals) > 1:
                puts(yellow("Warning: you selected more than one field while using `first=True` (order not guaranteed)"))
            res = vals[0]

    if cache:
        env['tabat_hash_cache'][tabat_hash_cache] = res
        puts("The result of the last psql query has been cached")
    return res

DigOutputLine = namedtuple('DigOutputLine', 'name ttl claz type content'.split())

def dig_query(domain, type=None, server=None, quiet=False):
    type = 'ANY' if type is None else type.upper()
    server = '' if server is None else '@' + server
    records = run('dig {0} +noall +answer +nocomments +nostats +retry=3 {1} {2}'.format(server, domain, type), quiet=quiet)
    lines = [DigOutputLine(*line.split(None, 4)) for line in records.splitlines()]
    if type != 'ANY':
        lines = [line for line in lines if line.type.upper() == type]
    return lines

def has_A_record(host):
    with settings(host_string='odoo@master.odoo.com'):
        dns_id = run('''psql -tAq -d saas_master -c "SELECT id FROM records WHERE name = '{0}' AND type = 'A'"'''.format(host), quiet=True, pty=False)

    return bool(dns_id)

def get_odoo_servers():
    sql = "SELECT public_ip, name, string_to_array(mode, ' ') as mode FROM saas_server"
    servers = psql_query('saas_master', sql, host_string='odoo@master.odoo.com', cache=True)
    return servers

#----------------------------------------------------------
# Setup
#----------------------------------------------------------

@task
@as_('root')
def service_reload(setup=False):
    """Reload static services: nginx, postgres, munin-node, fail2ban, openerp-smtpd"""
    fabtools.service.reload('nginx')
    fabtools.service.reload('postgresql')
    fabtools.service.reload('ssh')
    fabtools.service.restart('fail2ban')
    fabtools.service.restart('openerp-smtpd')
    if not setup:
        restart_munin()

@task
@as_('root')
def restart_munin():
    fabtools.service.restart('munin-node')

def setup_tune2fs():
    with warn_only():
        for dev in 'sda1 md1 md2 md3 vda1'.split():
            run('tune2fs -m0 -c0 -i0 /dev/' + dev)

def p2_dependencies():
    """ returns (debs, pip_packages) """
    debs = """
        python-dev
        python-babel
        python-dateutil
        python-decorator
        python-docopt
        python-docutils
        python-feedparser
        python-gdata
        python-geoip
        python-gevent
        python-html2text
        python-jinja2
        python-ldap
        python-libxslt1
        python-lxml
        python-mako
        python-markdown
        python-matplotlib
        python-mock
        python-ofxparse
        python-openid
        python-passlib
        python-pil
        python-pip
        python-psutil
        python-psycogreen
        python-psycopg2
        python-pychart
        python-pydot
        python-pyparsing
        python-pypdf
        python-pypdf2
        python-reportlab
        python-requests
        python-setproctitle
        python-simplejson
        python-tz
        python-unittest2
        python-vatnumber
        python-vobject
        python-webdav
        python-werkzeug
        python-xlrd
        python-xlsxwriter
        python-xlwt
        python-yaml
        python-zsi
    """.split()

    pips = """
        fabtools
        geoip2
        ovh
        phonenumbers
        python-slugify
        sphinx-patchqueue
        suds
        ebaysdk
        num2words==0.5.4
    """.split()
    return (debs, pips)

def p3_dependencies():
    """ returns (debs, pip_packages) """
    debs = """
        python3-dev
        python3-babel
        python3-dateutil
        python3-decorator
        python3-docopt
        python3-docutils
        python3-feedparser
        python3-geoip
        python3-gevent
        python3-html2text
        python3-jinja2
        python3-lxml
        python3-mako
        python3-markdown
        python3-matplotlib
        python3-mock
        python3-ofxparse
        python3-openid
        python3-passlib
        python3-pil
        python3-pip
        python3-psutil
        python3-psycopg2
        python3-pydot
        python3-pyparsing
        python3-pypdf2
        python3-reportlab
        python3-requests
        python3-setproctitle
        python3-simplejson
        python3-tz
        python3-unittest2
        python3-vatnumber
        python3-werkzeug
        python3-xlrd
        python3-xlsxwriter
        python3-yaml
    """.split()

    # non existant and/or unused P3 deps removed: python-gdata python-libxslt1 python3-webdav python-zsi
    # P3 package not packaged in xenial (-> pips): psycogreen vobject xlwt
    pips = """
        ebaysdk
        fabtools
        geoip2
        ovh
        num2words==0.5.4
        phonenumbers
        psycogreen
        pyldap
        pyOpenSSL
        python-slugify
        sphinx-patchqueue
        suds-jurko
        vobject
        xlwt
    """.split()
    return (debs, pips)

@task
@as_('root')
def setup_packages():
    """Install/Update debian and python packages needed"""
    codename = run('lsb_release -cs').strip()
    uninstall = """mlocate xinetd locate wkhtmltopdf whoopsie""".split()
    # local packages repo
    sio = io.BytesIO(b"deb http://nightly.openerp.com/deb/%s ./" % codename)
    put(sio, '/etc/apt/sources.list.d/odoo.list')

    sio = io.BytesIO(b"Package: nginx\nPin: origin nightly.openerp.com\nPin-Priority: 1001")
    put(sio, '/etc/apt/preferences.d/odoo')

    run('add-apt-repository -y ppa:maxmind/ppa')    # for geoipupdate

    base_debs = """
antiword
bsd-mailx
curl
fabric
fail2ban
fonts-freefont-ttf
fonts-khmeros-core
file
geoipupdate
git
graphviz
htop
jq
less
libdbd-pg-perl
libev-dev
libevent-dev
libfreetype6-dev
libjpeg8-dev
liblchown-perl
libldap2-dev
libpq-dev
libsasl2-dev
libtiff-dev
libwww-perl
libxml2-dev
libxslt1-dev
lsb-release
lsof
make
mc
moreutils
mosh
msmtp
msmtp-mta
munin
munin-node
munin-plugins-core
munin-plugins-extra
ncdu
nginx
npm
openntpd
p7zip-full
pg-activity
poppler-utils
postgresql
postgresql-contrib
psmisc
rsnapshot
rsync
screen
sudo
tree
unzip
uptimed
vim
wkhtmltox
zip
""".split()
    fabtools.deb.uninstall(uninstall, purge=True, options=['--quiet'])
    fabtools.deb.update_index()
    run("apt-get upgrade -y --force-yes")

    p2_debs, p2_pips = p2_dependencies()
    p3_debs, p3_pips = p3_dependencies()

    debs = base_debs + p2_debs + p3_debs
    fabtools.deb.install(debs, options=['--force-yes', '--ignore-missing'])

    # NOTE libevent-dev is required by gevent. /!\ version 1.0 of gevent will require libev-dev (and cython)
    # run('pip install cython -e git://github.com/surfly/gevent.git@1.0rc2#egg=gevent')

    # fabtools.python.install(python_pip, upgrade=False)
    run("pip2 install -q {}".format(' '.join(p2_pips)))
    run("pip3 install -q {}".format(' '.join(p3_pips)))

    # Nodejs
    run("ln -sf /usr/bin/nodejs /usr/bin/node")
    run("npm install -g less less-plugin-clean-css")


@task
@as_('root')
def setup_fonts():
    ensure_xz_file('/usr/share/fonts/truetype/Sun-ExtA.ttf', 'http://nightly.openerp.com/fonts/Sun-ExtA.ttf.xz')

    # Free Code39 fonts, from http://www.barcodesinc.com/free-barcode-font/ :
    #  - font-family "Free 3 of 9", includes basic Code39
    ensure_xz_file('/usr/share/fonts/truetype/FREE3OF9.TTF', 'http://nightly.openerp.com/fonts/FREE3OF9.TTF.xz')
    #  - font-family "Free 3 of 9 Extended", includes ASCII extension of Code39 too
    ensure_xz_file('/usr/share/fonts/truetype/FRE3OF9X.TTF', 'http://nightly.openerp.com/fonts/FRE3OF9X.TTF.xz')

    # reportlab fonts
    rl_fontpath = run('''python -c "import os,reportlab;print os.path.join(os.path.dirname(reportlab.__file__), 'fonts')"''')
    fabtools.require.directory(rl_fontpath)
    with cd(rl_fontpath):
        fabtools.require.file('pfbfer.zip', url='http://www.reportlab.com/ftp/pfbfer.zip')
        run('unzip -qn pfbfer.zip')


@task
@as_('root')
def update_pkg(pkg, silent=False):
    silent = str2bool(str(silent))
    run('apt-get update -qq')
    q = '-qq' if silent else '-q'
    run('DEBIAN_FRONTEND=noninteractive apt-get install --assume-yes --force-yes {0} {1}'.format(q, pkg))

@task
@as_('root')
def update_pip(*pkg):
    fabtools.python.install(pkg, upgrade=True)

def rsync_files():
    sudo('rsync -rtlE %s/custom/setup/odoo/ %s/' % (ODOO_SRC_PATH, HOME))
    sudo('ln -sf {0}/custom/default/saas_worker/metabase.py {1}/bin/oe-meta'.format(ODOO_SRC_PATH, HOME))

    if USER == 'odoo':
        run('/home/odoo/bin/rsync_symlink.py /home/odoo/src/custom/setup/etc /etc')
        run('/home/odoo/bin/rsync_symlink.py /home/odoo/src/custom/setup/usr /usr')

        run("find /etc/postgresql -name 'postgresql.local.conf' -type l -delete")
        run("find /etc/postgresql -name 'main' -type d -exec touch '{}/postgresql.local.conf' ';' -exec chown postgres:postgres '{}/postgresql.local.conf' ';'")

        run('rsync -rtlE /home/odoo/src/custom/setup/etc_copy/ /etc/')
        run('rsync -rtlE /home/odoo/src/custom/setup/root/ /root/')
        run('rsync -rtlE /home/odoo/src/custom/setup/odoo/.ssh/ /root/.ssh/')

        run('sysctl -e -p')    # reload /etc/sysctl.conf

        run('chmod 775 /etc/nginx/odoo-static')
        run('chown root:odoo /etc/nginx/odoo-static')
        run('chown postgres:postgres /etc/postgresql/*/main/postgresql.conf')

    run('chmod 440 /etc/sudoers.d/*')
    run('chmod 700 %s/.ssh %s/.ssh/* /root/.ssh /root/.ssh/*' % (HOME, HOME))
    run('chmod 755 /home %s %s %s/custom' % (HOME, ODOO_SRC_PATH, ODOO_SRC_PATH))
    if has_systemd():
        run('systemctl daemon-reload')

@task
@as_('root')
def setup_user():
    if not fabtools.user.exists(USER):
        if USER != 'odoo':
            abort('user %r does not exists' % USER)
            return
        fabtools.user.create('odoo', home=HOME, shell='/bin/bash', uid='10000')

    sudo('mkdir -p ' + ODOO_SRC_PATH)
    with cd(ODOO_SRC_PATH):
        if not fabtools.files.is_dir('custom'):
            if not fabtools.files.is_file('{0}/.ssh/id_rsa'.format(HOME)):
                # this is bootstrap => need to have the correct ssh key and config
                sudo('mkdir -p {0}/.ssh'.format(HOME))
                put(os.path.join(HERE, 'odoo/.ssh/id_rsa*'), '{0}/.ssh/'.format(HOME))
                sudo('ssh-keyscan -t rsa -H github.com >> {0}/.ssh/known_hosts'.format(HOME))
                run('chmod 700 {0}/.ssh {0}/.ssh/*'.format(HOME))
                run('chown {0}:{0} {1}/.ssh/*'.format(USER, HOME))
            sudo('git clone git@github.com:odoo/internal.git custom')
        else:
            _update('custom')

        if not fabtools.files.is_dir('migration'):
            sudo('git clone git@github.com:odoo/saas-migration.git migration')
        else:
            _update('migration')

    sudo('mkdir -p {0}/log'.format(HOME))
    fabtools.require.directory('/home/bak/odoo')    # only, user odoo, other users are not backuped on test


@task
@as_('root')
def setup_services():
    rsync_files()
    if not fabtools.systemd.is_running('nginx'):
        fabtools.systemd.start('nginx')
    fabtools.systemd.enable('nginx')
    fabtools.systemd.enable('openerp-smtpd')


def setup_locales():
    langs = """
ar_AE ar_BH ar_DZ ar_EG ar_IQ ar_JO ar_KW ar_LB ar_LY ar_MA ar_OM ar_QA ar_SA ar_SD ar_SY ar_TN ar_YE bg_BG bs_BA ca_ES cs_CZ de_DE en_GB en_US es_AR es_ES et_EE fi_FI fr_BE fr_FR hr_HR hu_HU id_ID it_IT nl_BE nl_NL pl_PL pt_BR pt_PT ro_RO ru_RU sl_SI sq_AL sv_SE tr_TR uk_UA zh_CN zh_TW
    """.split()

    new_langs = []
    with warn_only():
        for l in langs:
            # run('grep {0}.UTF-8 /var/lib/locales/supported.d/local || locale-gen {0}.UTF-8'.format(l))
            g = run('grep -q {0}.UTF-8 /var/lib/locales/supported.d/local'.format(l))
            if g.failed:
                new_langs.append(l)

    if new_langs:
        run('locale-gen ' + ' '.join(l + '.UTF-8' for l in new_langs))

def setup_postgres(version='9.5'):
    datadir = '/home/postgres/%s/main' % version
    if fabtools.files.is_dir(datadir):
        return
    fabtools.require.directory('/home/postgres/%s/main' % version)
    run('chown -R postgres:postgres /home/postgres')
    sudo('/usr/lib/postgresql/%s/bin/initdb --locale=en_US.UTF-8 --lc-collate=C %s' % (version, datadir), user='postgres')
    fabtools.service.start('postgresql')
    sudo('''psql -c "CREATE USER root WITH SUPERUSER PASSWORD 'root' "''', user='postgres')
    sudo('''psql -c "CREATE USER odoo WITH LOGIN PASSWORD 'odoo' "''', user='postgres')
    sudo('''psql -c "CREATE USER replicant WITH LOGIN REPLICATION PASSWORD 'eratEwUlCijMiptedsAsUdKetphoick6' "''', user='postgres')

@task
@as_('root')
def setup_metabase():
    """Create or update schema of `meta` database"""
    META = 'meta'
    with settings(sudo_user=env.user):
        if not fabtools.postgres.database_exists(META):
            fabtools.postgres.create_database(META, owner='root')

    with settings(sudo_user='odoo'), cd('~'), shell_env(PGOPTIONS='--client-min-messages=warning'):
        sudo('psql -Xq -d {0} -f {1}/custom/setup/meta.sql'.format(META, ODOO_SRC_PATH))

def setup_munin():
    with warn_only():
        with cd('/etc/munin/plugins'):
            run('rm fw_forwarded_local if_dummy* if_eth1 if_ip6* if_tun* postgres_* proc_pri smart_*')
            run('ln -sf /usr/share/munin/plugins/{diskstats,iostat*,nginx*,vmstat} .')  # for some reason not pre-installed in Ubuntu 12.10+

    cache = '/var/cache/munin/www'
    fabtools.require.directory(cache)
    run('chown -R munin:munin ' + cache)

@task
@as_('root')
def setup():
    """Setup or update packages services and odoo"""
    setup_tune2fs()
    setup_packages()
    setup_fonts()
    fabtools.require.service.stopped('postgresql')
    setup_user()
    setup_services()
    setup_locales()
    setup_postgres()
    fabtools.require.service.restarted('postgresql')
    setup_metabase()
    setup_munin()
    update_geoip(restart_openerp=False)
    service_reload(setup=True)

#----------------------------------------------------------
# openerp series handling
#----------------------------------------------------------

@as_('root')
def _create_initd_script(branch):
    ctx = {
        'branch': branch
    }

    sudo('ln -sf {0}/bin/openerp {0}/bin/openerp-{1}'.format(HOME, branch))

    def _upload_template(template, target, mode):
        upload_template(os.path.join(HERE, template), target, ctx, backup=False, mode=mode)

    if USER == 'odoo':
        service_name, service_path = branch2service(branch)
        if has_systemd():
            # systemd
            _upload_template('unit_openerp.tpl', service_path, '0644')
            run('systemctl daemon-reload')
            fabtools.systemd.enable(service_name)
        else:
            # SysV init
            _upload_template('initd_openerp.tpl', service_path, '0755')
            run('update-rc.d {0} defaults'.format(service_name))


@task
@as_('root')
def checkout(branch_nick):
    """Checkout a saas branch and create init.d script.  i.e: checkout:saas-1"""

    def co(repo, url, condition=lambda: True):
        with cd(ODOO_SRC_PATH + '/' + repo):
            if not fabtools.files.is_dir(branch_nick) and condition():
                cmd = 'git ls-remote --heads {0} "{1}"'.format(url, branch_nick)
                if local(cmd, capture=True).strip():
                    sudo('git clone -q --branch {0} --single-branch {1} {0}'.format(branch_nick, url))

    sudo('mkdir -p {0}/{{odoo,themes,enterprise}}'.format(ODOO_SRC_PATH))
    co('odoo', 'git@github.com:odoo/odoo.git')
    co('themes', 'git@github.com:odoo/design-themes.git',
       lambda: branch_nick != '8.0')
    co('enterprise', 'git@github.com:odoo/enterprise.git')

    sudo('mkdir -p {0}/filestore'.format(HOME))
    # resymlink
    src_path = None
    for product in 'OpenERP Odoo'.split():
        sudo('mkdir -p {0}/.local/share/{1}/sessions'.format(HOME, product))
        sudo('ln -sfT {0}/filestore {0}/.local/share/{1}/filestore'.format(HOME, product))
        path = '{0}/odoo/{1}/{2}'.format(ODOO_SRC_PATH, branch_nick, product.lower())
        if fabtools.files.is_dir(path):
            src_path = path

    assert src_path

    sudo('ln -sfT {0}/migration {1}/addons/base/maintenance'.format(ODOO_SRC_PATH, src_path))
    sudo('ln -sfT {0}/filestore {1}/filestore'.format(HOME, src_path))

    _create_initd_script(branch_nick)

def _repos(branch_nick):
    return (d for d in ['odoo', 'themes', 'enterprise']
            if fabtools.files.is_dir(os.path.join(ODOO_SRC_PATH, d, branch_nick)))

@task
@as_('root')
def remove_branch(branch_nick, check_stoped=True):
    """remove a given branch"""

    if USER == 'odoo':
        dbcount = sudo('{0}/bin/oe-meta count {1}'.format(HOME, branch_nick))
        if int(dbcount):
            abort('There is still {0} databases listed in metabase'.format(dbcount))
            return

        service_name, service_path = branch2service(branch_nick)

        check_stoped = str2bool(str(check_stoped))
        if check_stoped:
            if has_systemd():
                running = fabtools.systemd.is_running(service_name)
            else:
                running = not run(service_path + ' status').strip().endswith('stopped')

            if running:
                abort('odoo {0} still running on {1}'.format(branch_nick, env.host_string))
                return

        if has_systemd():
            fabtools.systemd.disable(service_name)
        else:
            run('update-rc.d -f {0} remove'.format(service_name))

        run('rm -f {0}'.format(service_path))

    # NOTE keep logs
    sudo('rm -f {0}/bin/openerp-{1} {0}/config/openerp-{1}.sh'.format(HOME, branch_nick))

    # NOTE: we cannot use bash curly-brace expansion as it doesn't expand 1-item lists
    r = ' '.join('{0}/{1}/{2}'.format(ODOO_SRC_PATH, b, branch_nick) for b in _repos(branch_nick))
    if r:
        sudo('rm -rf ' + r)

@task
@as_('root')
def update(branch_nick, restart_openerp=True):
    """Update a saas branch (server, addons, web) and restart openerp.  i.e: update:saas-1"""
    friday_check('update ' + branch_nick)
    repos = OrderedDict.fromkeys(_repos(branch_nick))
    for repo in repos:
        repos[repo] = _update('{0}/{1}'.format(repo, branch_nick))
        if not repos[repo]:
            puts(red('Conflicts occured during update of {0}, branch {1}. Not restarting.'.format(repo, branch_nick)))
    restart_openerp = all(repos.values()) and str2bool(str(restart_openerp))
    if restart_openerp:
        restart(branch_nick)

@task
@as_('root')
def update_custom(restart_openerp=True):
    """Update custom branch and restart all running openerp servers"""
    friday_check('update custom')
    restart_openerp = str2bool(str(restart_openerp))
    _update_custom()
    if not restart_openerp:
        return
    for b in all_branches():
        restart(b)

def _update_custom():
    """update custom and migration branches"""
    _update('custom')
    _update('migration')
    # == deprecated ==
    # rsync_files()
    # for b in all_branches(True):
    #     _create_initd_script(b)

def _update(repo):
    path = ODOO_SRC_PATH + '/' + repo
    if not fabtools.files.is_dir(path):
        puts(yellow('skipping non-existing {0}'.format(path)))
        return True

    puts(blue('update {0}'.format(repo)))

    is_frontend = any(env.host_string in r for r in env.roledefs['of'])
    if not is_frontend: # recursive chown on NFS mounts is Slow(TM)
        run("chown -R {0}:{0} {1}".format(USER, path))

    with cd(path):
        sudo('git fetch --quiet --prune')
        sudo('git rebase --quiet --autostash')
        with hide('status', 'commands'):
            no_conflicts = sudo('git diff --diff-filter=U --no-patch --exit-code').succeeded

    return no_conflicts

@task
@as_('root')
def update_all():
    """Update custom and all running openerp's"""
    friday_check('update all running openerp')
    _update_custom()
    for b in all_branches():
        update(b)

@task
@as_('root')
def update_lb(reload_ha=True):
    """ Update load balancer """
    friday_check('update custom')
    reload_ha = str2bool(str(reload_ha))
    _update('custom')
    rsync_files()

    ssl_dir = '/etc/ssl'
    pem_dir = ssl_dir + '/pem'
    fabtools.require.directory(pem_dir)
    ext = 'crt'

    def file_exists(path):
        with settings(hide('running', 'warnings'), warn_only=True):
            return run('[ -e "{0}" ]'.format(path)).succeeded

    for crt_path in run("ls -1p {0}/_*.{1}".format(ssl_dir, ext)).splitlines():
        crt = crt_path.rpartition('/')[-1]
        basename = crt[:-len(ext)]
        key = basename + 'key'
        key_path = '{0}/{1}'.format(ssl_dir, key)
        # ignore broken symlinks
        if file_exists(key_path) and file_exists(crt_path):
            pem = basename + 'pem'
            # concat the crt and the key with awk to ensure we have a \n
            # between the two.
            run("awk '1' {0}/{2} {0}/{3} > {1}/{4}".format(ssl_dir, pem_dir, crt, key, pem))

    if reload_ha:
        with settings(hide('running', 'warnings'), warn_only=True):
            if not run('haproxy -c -f /etc/haproxy/haproxy.cfg').succeeded:
                abort(red('Invalid haproxy config!'))
        fabtools.systemd.reload('haproxy')

@task
@as_('root')
def update_prod(restart_odoo=True):
    friday_check('update prod')
    restart_odoo = str2bool(str(restart_odoo))
    for ob in env.roledefs['ob']:
        with settings(host_string=ob):
            _update('custom')
            for b in all_branches():
                update(b, restart_odoo)
    if restart_odoo:
        restart_prod(False)

@task
@as_('root')
def restart_prod(ob=True):
    friday_check('restart prod')
    s = env.roledefs['of']
    if ob:
        s = s + env.roledefs['ob']      # no +=

    for of in s:
        with settings(host_string=of):
            for b in all_branches():
                restart(b)

@task
@as_('root')
def restart(branch_nick):
    """Restart an openerp server.  i.e: restart:saas-1"""
    _service_action(branch_nick, 'restart')

@task
@as_('root')
def stop(branch_nick):
    """Stop an openerp server.  i.e: stop:saas-1"""
    _service_action(branch_nick, 'stop')

def _service_action(branch_nick, action):
    service_name, service_path = branch2service(branch_nick)

    if USER != 'odoo':
        puts(yellow('NOT %sING %s for user %s. please restart it manually' % (action.upper(), branch_nick, USER)))
    elif not fabtools.files.is_file(service_path):
        puts(yellow('service {0} missing: skipping'.format(branch_nick)))
    else:
        if has_systemd():
            getattr(fabtools.systemd, action)(service_name)
        else:
            run('{0} {1}'.format(service_path, action))

@task
@as_('root')
def status(branch_nick):
    """Show status of odoo server"""
    service_name, service_path = branch2service(branch_nick)

    if USER != 'odoo':
        puts(yellow('unknow status of %s for user %s.' % (branch_nick, USER)))
    elif not fabtools.files.is_file(service_path):
        puts(yellow('service {0} missing: skipping'.format(branch_nick)))
    else:
        if has_systemd():
            run('systemctl status {0}'.format(service_name))
        else:
            run('{0} status'.format(service_path))

@task
@as_('root')
def restart_all():
    """restart all runnning openerp's"""
    for b in all_branches():
        restart(b)

@task
@as_('root')
def missing(branch_nick=None):
    """Show missing commits in respositories"""

    gitmissing = "git --git-dir={repo}/.git fetch --quiet && git --git-dir={repo}/.git log --decorate --oneline --graph --color ..origin/{branch} | sed '/^$/d'".format

    with hide('running'):
        if branch_nick is None:
            with hide('stdout'):
                branch_nick = sudo('~/bin/oe-meta last-branch')

        with cd(ODOO_SRC_PATH):
            for p in ['custom', 'migration']:
                puts(green(p))
                sudo(gitmissing(repo=p, branch='master'))

            path = 'odoo/' + branch_nick
            if not fabtools.files.is_dir(path):
                puts(yellow('skipping non-existing {0}'.format(path)))
            else:
                puts(green(branch_nick))
                sudo(gitmissing(repo=path, branch=branch_nick))

@task
@as_('root')
def update_geoip(restart_openerp=True):
    run('geoipupdate')
    restart_openerp = str2bool(str(restart_openerp))
    if restart_openerp:
        restart_all()
#----------------------------------------------------------
# openerp db management
#----------------------------------------------------------

@task
def clean_dumps():
    run('rm -f /tmp/oe-saas-dump/*')

@task
@as_('root')
def migrate(db, to=None):
    """Migrage a database to a new version. Simply call ~odoo/bin/oe-meta migrate"""
    if to is None:
        to = ''
    sudo('{0}/bin/oe-meta migrate {1} {2}'.format(HOME, db, to))

@task
@as_('root')
def list_dbuuids(version):
    """list emails of admins of databases. """
    MODES = ('paid', 'trial')   # not drop
    NOT_STATUSES = ('cancel', 'close', 'unsubscribed',
                    'maintenance', 'maintenance_mig', 'maintenance_block')

    with hide('commands'):
        with cd('~'):
            q = "SELECT uuid FROM databases WHERE coalesce(status,'') not in %s and mode in %s and port=%d"
            q = q % (NOT_STATUSES, MODES, get_port(version))
            lst = sudo('psql -X --pset=pager=off -tAq -d meta -c "%s"' % q)
    dbuuids = env.setdefault('dbuuids', [])
    dbuuids.extend(lst.strip().split())

    if env.host_string == env.all_hosts[-1]:
        domain = "[('subscription_id', '!=', False), ('db_uuid', 'in', %r)]" % (dbuuids,)
        fastprint(domain, end="\n")

@task
@as_('root')
def ensure_base_version(version=''):
    queries = [
        # "UPDATE ir_module_module SET latest_version = '7.0.1.3' WHERE latest_version = '7.0.7.0.1.3';",
        "UPDATE ir_cron SET nextcall = CURRENT_TIMESTAMP(0) at time zone 'UTC' WHERE model = 'publisher_warranty.contract';",
    ]
    lst = sudo('{0}/bin/oe-meta list -m trial -m paid {1}'.format(HOME, version)).strip().split()
    lst.sort()
    #lst.pop(lst.index('ssu70test'))
    l = len(lst)
    with cd('~'):
        for i, db in enumerate(lst):
            pc = int(100.0 * (i + 1) / l)
            puts('%d/%d (%d%%)' % (i + 1, l, pc))
            for query in queries:
                sudo('psql -d {0} -c "{1}"'.format(db, query))

@task
def move_db(db, to, no_crud_since=0, force=False):
    """Move database to another host. i.e: move_db:agrolait,eu16a,no_crud_since=<min>,force=<bool>
        no_crud_since=<min>: ensures that no CRUD activity (http POSTs) has been made since the `n` last minutes.
    """
    host = db + '.odoo.com'
    _move_db(db, host, to, no_crud_since=no_crud_since, force=force)

def gethostbyname(h):
    return local('dig +short @master.odoo.com {0}'.format(h), capture=True).strip()

def _move_db(db, from_, to, abort=abort, no_crud_since=0, force=False):
    trial_servers = [t.partition('@')[-1] for t in env.roledefs['trial']]
    if '.' not in to:
        to = dict((t.split('.')[0], t) for t in trial_servers).get(to, to)
    assert to in trial_servers
    assert '.' not in db

    host = db + '.odoo.com'

    if gethostbyname(host) == gethostbyname(to):
        abort('database {0} is already on {1}'.format(db, to))
        return

    # sometimes we have databases that exist but don't have a DNS
    # record anymore on master.odoo.com, those should be skipped
    # because what'll happen is master will return the IP of another
    # server (to show 'this db doesn't exist, do you want to create
    # it?'), and the issue is that fab won't have access to that
    # server (we can only get on ourselves, master and eu11 from here)
    # and then the fab task will get stuck (all night) because it'll
    # prompt for an ssh password.
    if not force and not has_A_record(host):
        puts(yellow('Skipping %s because it doesn\'t have an A record on master.odoo.com' % host))
        return

    # there are certain ways people can create database names that end
    # with "-", which leads to invalid subdomains like
    # abc-.odoo.com. Skip those, because certain tools won't be able
    # to resolve those types of subdomains.
    if not force and db.endswith("-"):
        puts(yellow('Skipping %s because it ends with "-"' % host))
        return

    env.forward_agent = True
    with settings(host_string='odoo@' + from_):
        if int(no_crud_since):
            # This is a naive heuristic for CRUD activity detection: checking log files for http POST
            last_log = run("cat log/openerp-*.log | grep -E '%s werkzeug:.+\"POST /' | grep -vE '(/calendar/notify|/longpolling/poll)' | tail -n1" % db).strip()
            if last_log:
                date_fmt = '%Y-%m-%d %H:%M:%S'
                server_date = datetime.datetime.strptime(run('date --utc "+%s"' % date_fmt, quiet=True).strip(), date_fmt)
                log_date = datetime.datetime.strptime(last_log.split(',')[0], date_fmt)
                last_activity = (server_date - log_date).total_seconds() // 60  # in minutes
                if not force and last_activity < int(no_crud_since):
                    puts(yellow('Last CRUD activity found %d minute(s) ago... Skipping.' % last_activity))
                    return

        # safety check for custom smtp
        reverse_dns = run('dig +short -x $(dig +short %s)' % to)
        if 'googleusercontent' in reverse_dns:
            custom_mail_server = int(run('''psql -tAq -d {0} -c "SELECT count(*) FROM ir_mail_server WHERE name != 'localhost' AND smtp_port = 25"'''.format(db)))
            if custom_mail_server:
                abort(red('Database {0} uses custom SMTP server and cannot be hosted on {1} due to Google outgoing ports restrictions. Please move it elsewhere'.format(db, to)))
                return

        dump = run('~/bin/oe-meta dumps {0}'.format(db), quiet=True)
        if not dump:
            abort(red('unknown database {0} on {1}'.format(db, from_)))
            return

        # Could use `oe-meta get -j` but this would be one more ssh connection (enouf is enouf)
        zdata = base64.urlsafe_b64decode(dump.encode() + b'=' * (-len(dump) % 4))
        jrecord = json.loads(zlib.decompress(zdata).decode())
        domains = [dom for dom in jrecord['hosts'] if not dom.endswith(('.odoo.com', '.openerp.com'))]

        if domains:
            puts(white("The external domains %s are registrered for the database %s" % (' and '.join(domains), db)))
            servers_ips = [serv['public_ip'] for serv in get_odoo_servers()]
            servers_names = [serv['name'] for serv in get_odoo_servers()]
            dns_conflicts = []
            for dom in domains:
                for line in dig_query(dom, server='8.8.8.8'):
                    conditions = (
                        line.type == 'A' and line.content.strip('.') in servers_ips,
                        line.type == 'MX' and any([server.lower() in line.content.lower() for server in servers_names])
                    )
                    if any(conditions):
                        dns_conflicts.append(line)
            if not force and dns_conflicts:
                puts(yellow("Cannot move %s because of the following DNS misconfiguration:" % host))
                for line in dns_conflicts:
                    msg = " - the domain %s DNS record (%s) points to an Odoo server (%s)"
                    puts(yellow(msg % (line.name, line.type, line.content)))
                return

        run('~/bin/oe-meta block {0}'.format(db))
        for t in range(6):
            if t:
                puts('sleep %d' % (t,))
            time.sleep(t)
            cnx_cnt = int(run('''psql -tAq -d postgres -c "SELECT count(1) FROM pg_stat_activity WHERE datname='{0}' and state!='idle';"'''.format(db)))
            if cnx_cnt == 0:
                break
        else:
            run('~/bin/oe-meta unlock {0}'.format(db))
            abort('database {0} has active connections. aborting'.format(db))
            return

        # we sometimes use a pg_waiter_holder view to examine
        # performance related matters. it uses the hold_act pid
        # however, which is called pid in psql 9.1 and procpid in psql
        # 9.2. so to avoid issues during moving the db we just drop
        # those views.
        run('''psql -tAq -d {0} -c "DROP VIEW IF EXISTS pg_waiter_holder;"'''.format(db), warn_only=True, quiet=True)

        psql_version_cmd = '''psql -tAq -d postgres -c "SHOW server_version_num;"'''
        source_psql_version = int(run(psql_version_cmd))
        with settings(host_string='root@' + to):
            remote_psql_version = int(run(psql_version_cmd))

        compatible = True
        if source_psql_version >= 90500:
            if 90300 <= remote_psql_version < 90500:
                compatible = False
            elif remote_psql_version < 90512 <= source_psql_version:  # cf. CVE-2018-1058
                compatible = False

        if compatible:
            run('pg_dump -Fc --serializable-deferrable {0} | ssh -q root@{1} pg_restore -x -Fc -C -d postgres'.format(db, to))
        else:
            # Plain dump + get rid of REPLICA IDENTITY statements (for 9.3)
            with settings(host_string='root@' + to):
                remote_psql_version = run('createdb {0}'.format(db))
            run("""pg_dump -Fp --serializable-deferrable {0} | grep -v "REPLICA IDENTITY" | gzip -c | ssh -q root@{1} 'zcat - | psql -X --pset=pager=off -d {0}'""".format(db, to))

        fs = '/home/odoo/filestore/{0}/'.format(db)
        if fabtools.files.is_dir(fs):
            run('rsync -aqz --progress {0} odoo@{1}:{0}'.format(fs, to))

    with settings(host_string='odoo@' + to):
        run('~/bin/oe-meta loads {0}'.format(dump))

    with settings(host_string='odoo@master.odoo.com'):
        ttl = run('''psql -tAq -d saas_master -c "UPDATE records SET content='{0}' WHERE name in {1} AND type='MX';UPDATE records SET content=(SELECT public_ip FROM saas_server WHERE name='{0}') WHERE name in {1} AND type='A' RETURNING ttl;"'''.format(to, "('%s', '%s')" % (host, db + '.my.openerp.com')))

    with settings(host_string='odoo@' + from_):
        ttl = int(int(ttl.splitlines()[0]) * 3)
        run('''~/bin/oe-meta set -m drop -e "now() at time zone 'UTC' + interval '{1} seconds'" {0}'''.format(db, ttl))


@task
@as_('odoo')
@graceful_stop
def move_all_dbs(to, version='', nodrop=True, order='', max=0, no_crud_since=0, confirm=False):
    """usage: move_all_dbs:<to>,version=<version>,nodrop=<bool>,order=<order>,max=<max>,no_crud_since=<min>
       If nodrop is set to false, it will include "drop" databases.
       order allows to sort hows databases are processed. e.g: order='last_cnx_date desc'

       no_crud_since=<min>: ensures that no CRUD activity (http POSTs) has been made since the `n` last minutes.
    """
    assert len(env.hosts) == 1, "I don't know how fabric will behave if you use multiple hosts. do it one by one"
    if version:
        version = ' ' + version
    friday_check('move all%s databases' % version)
    with hide('stdout'):
        c = json.loads(run('cat ~/config/config.json'))
    if not version and 'new' in c['me']['mode'].split():
        abort('This server still accept new instances')
        return

    nodrop = str2bool(str(nodrop))
    modes = '-m trial -m paid'
    if not nodrop:
        modes += ' -m drop'
    if order:
        order = ' -o "%s"' % order
    with hide('stdout'):
        dbs = run('~/bin/oe-meta list ' + modes + version + order).splitlines()
    if max:
        dbs = dbs[0:int(max)]
    l = len(dbs)
    for i, db in check_stop(enumerate(dbs, 1)):
        pc = int(100.0 * i / l)
        msg = blue('move %s: %d/%d (%d%%)' % (db, i, l, pc))
        if confirm:
            if not _confirm(msg):
                continue
        else:
            puts(msg)
        _move_db(db, env.host, to, abort=warn, no_crud_since=no_crud_since)


@task
@as_('odoo')
def drop_dbs_older_than(laps='2 years', version=None, modes='drop trial paid', confirm=True):
    """usage: drop_dbs_older_than:'2 years',version=<version>,modes=<modes>,confirm=<bool>"""
    modes = ', '.join(["'%s'" % mode.strip() for mode in modes.split()])
    port = ''
    if version:
        port = ' AND port = %s' % run('oe-meta get-port %s' % version)
    sql = '''
        SELECT db.name AS dbname, (SELECT string_agg(dom.name, '\\n') FROM domains dom WHERE dom.db_id = db.id) AS domains,
               db.status, db.mode, db.create_date, db.last_cnx_date
        FROM databases db
        WHERE db.mode IN ({modes}) {port} AND last_cnx_date < now() - interval '{laps}'
        ORDER BY db.last_cnx_date
    '''.format(**locals())
    dbs = psql_query('meta', sql)
    if not dbs:
        abort(yellow("No database found"))

    import tabulate  # noqa
    print(tabulate.tabulate(dbs, headers='keys', tablefmt='pipe'))
    if not confirm or _confirm(red("Are you sure you want to drop those databases?"), default=False):
        for database in dbs:
            run('oe-meta set -m drop -e "now()" %s' % database['dbname'])


@task
@as_('root')
def update_dbuuids():
    with hide('status', 'commands'):
        dbs = sudo('{0}/bin/oe-meta list -m paid -m trial'.format(HOME)).strip().split()
    query = "SELECT value FROM ir_config_parameter WHERE key='database.uuid';"
    with cd('~'):
        l = len(dbs)
        for i, db in enumerate(dbs, 1):
            pc = int(100.0 * i / l)
            puts('update uuid %s: %d/%d (%d%%)' % (db, i, l, pc))
            out = sudo('psql -tA -d {0} -c "{1}"'.format(db, query), warn_only=True)
            if out.failed:
                continue
            out = out.strip()
            sudo('{0}/bin/oe-meta set -u {1} {2}'.format(HOME, out, db))

@task
def restore_db(db, trial_days=None, backup_file=False):
    """ restore_db:name[,15[,backup_file]]"""
    host = db + '.odoo.com'

    # NOTE cannot use gethostbyname due to typo wildcard response.
    mx_host = local('dig +short MX @master.odoo.com {0}'.format(host), capture=True).strip()
    if mx_host:
        abort("domain already exists")

    backup_servers_query = """
        SELECT name FROM (
            SELECT name, row_number() OVER (PARTITION BY backup_group_id ORDER BY random()) AS r
              FROM saas_server
             WHERE mode ilike 'bak%'
               AND backup_group_id is not null
        ) AS t WHERE r = 1;
    """
    with settings(host_string='odoo@master.odoo.com'):
        with hide('output', 'running'):
            backup_servers = run('''psql -tAq -d saas_master -c "{}"'''.format(backup_servers_query)).splitlines()
    # add legacy servers
    backup_servers += ['bak7.odoo.com', 'bak8.odoo.com', 'bak9.odoo.com']

    env.forward_agent = True
    last_backup = None
    for backup_server in backup_servers:
        with settings(host_string='root@' + backup_server):
            if backup_file:
                if exists(backup_file):
                    last_backup = backup_file
            else:
                # compatibity with both dump formats (with and without uuid + timestamp)
                last_backup_res = run('ls -1t /home/bak/odoo/*/*/home/odoo/bakdb/{0}.*.sql.gz | head -n1'.format(db)).strip()
                if last_backup_res[0] != "/":
                    last_backup_res = run('ls -1t /home/bak/odoo/*/*/home/odoo/bakdb/{0}.sql.gz | head -n1'.format(db)).strip()
                if last_backup_res[0] == "/":
                    last_backup = last_backup_res
            if last_backup:
                break

    if not last_backup:
        if backup_file:
            abort("Specified backup file does not exist")
        else:
            abort("No backup has been found for database '%s'." % (db))

    with settings(host_string='root@' + backup_server):
        if last_backup[0] == '/':
            # found on bakcup server
            bs = last_backup.split('/')

            dest_host = bs[5]
            actual_db = bs[-1].split('.')[0]

            run('scp {0} odoo@{1}:/tmp/{2}.sql.gz'.format(last_backup, dest_host, db))

            filestore = '/'.join(bs[:8] + ['filestore', actual_db])
            if fabtools.files.is_dir(filestore):
                run('rsync -aqz {0}/ odoo@{1}:/home/odoo/filestore/{2}/'.format(filestore, dest_host, db))

        else:
            # not found on backup server (too old?), search on every trial server
            bakarch = '/home/odoo/bakdb_archive/{0}.sql.gz'.format(db)
            for trial_server in env.roledefs['trial-only']:
                trial_server = trial_server.partition('@')[-1]
                with settings(host_string='odoo@' + trial_server):
                    if fabtools.files.is_file(bakarch):
                        dest_host = trial_server
                        run('cp {0} /tmp/{1}.sql.gz'.format(bakarch, db))
                        break
            else:
                abort('Cannot find backup for datbase {0}'.format(db))

        with settings(host_string='root@' + dest_host):
            run('createdb {0}'.format(db))
            run('psql -tAq -c "CREATE EXTENSION IF NOT EXISTS plpgsql" -d {0}'.format(db))
            run('psql -tAq -c "CREATE EXTENSION IF NOT EXISTS unaccent" -d {0}'.format(db))
        with settings(host_string='odoo@' + dest_host):
            run('zcat /tmp/{0}.sql.gz | psql -q -o /dev/null {0}'.format(db))
            if backup_file:
                dbuuid = str(uuid4())
                run('''psql -tAq -d {0} -c "UPDATE ir_config_parameter SET value='{1}' WHERE key='{2}'"'''.format(db, dbuuid, 'database.uuid'))
                has_oauth = run('''psql -tAq -d {0} -c "SELECT EXISTS(SELECT 1 FROM pg_tables WHERE tablename = 'auth_oauth_provider')"'''.format(db))
                if has_oauth == 't':
                    run('''psql -tAq -d {0} -c "UPDATE auth_oauth_provider SET client_id='{1}' WHERE id = (SELECT res_id FROM ir_model_data WHERE model='auth.oauth.provider' AND name='provider_openerp')"'''.format(db, dbuuid))
            else:
                dbuuid = run('''psql -tAq -d {0} -c "SELECT value FROM ir_config_parameter WHERE key='{1}'"'''.format(db, 'database.uuid')).strip()
            create_date = run('''psql -tAq -d {0} -c "SELECT value FROM ir_config_parameter WHERE key='{1}'"'''.format(db, 'database.create_date')).strip()
            series = run(r'''psql -tAqX -d {0} -c "select replace((regexp_matches(latest_version, '^\d+\.0|^saas~\d+\.\d+|saas~\d+'))[1], '~', '-') from ir_module_module where name='base'"'''.format(db)).strip()

    # create dns entries
    with settings(host_string='odoo@master.odoo.com'):
        run('''psql -tAq -d saas_master -c "INSERT INTO records(create_date, write_uid, name, content, type, ttl, prio, origin, domain_id) VALUES (now(), 1, '{host}', '{server}', 'MX', 600, 10, 'saas', 1), (now(), 1, '{host}', (SELECT public_ip FROM saas_server WHERE name='{server}'), 'A', 600, 0, 'saas', 1);"'''.format(host=host, server=dest_host))
        run('''psql -tAq -d saas_master -c "INSERT INTO records(create_date, write_uid, name, content, type, ttl, prio, origin, domain_id) VALUES (now(), 1, '{host}', '{server}', 'MX', 600, 10, 'saas', 1), (now(), 1, '{host}', (SELECT public_ip FROM saas_server WHERE name='{server}'), 'A', 600, 0, 'saas', 1);"'''.format(host=db + '.my.openerp.com', server=dest_host))

    # update metabase
    with settings(host_string='odoo@' + dest_host):
        args = {
            'm': 'paid',
            'p': get_port(series, with_sudo=False),
            'c': '"%s"' % create_date,
            'H': '"%s"' % host,
            'u': '"%s"' % dbuuid,
        }

        if trial_days:
            exp_date = (datetime.datetime.now() + relativedelta(days=int(trial_days), microsecond=0)).isoformat()
            args.update(m='trial', e='"%s"' % exp_date)

        args_str = ' '.join('-%s %s' % x for x in args.items())
        run('~/bin/oe-meta set -C {0} {1}'.format(args_str, db))

@task
def restore_migrated_db(db, server, upgrade_url, trial_days=30):
    """ restore_migrated_db:name, server, upgrade_url, [trial_days=30]"""
    host = db + '.odoo.com'
    echo = partial(puts, show_prefix=False)

    # NOTE cannot use gethostbyname due to typo wildcard response.
    mx_host = local('dig +short MX @master.odoo.com {0}'.format(host), capture=True).strip()
    if mx_host:
        abort("domain already exists")

    with settings(host_string='root@' + server):
        run('createdb {0}'.format(db))
        run('psql -tAq -c "CREATE EXTENSION IF NOT EXISTS plpgsql" -d {0}'.format(db))
        run('psql -tAq -c "CREATE EXTENSION IF NOT EXISTS unaccent" -d {0}'.format(db))
    with settings(host_string='odoo@' + server), temporary_directory() as tmp_dir:
        run('wget -q {0} -O {1}/upgraded_db.zip'.format(upgrade_url, tmp_dir))
        run('unzip -qq {0}/upgraded_db.zip -d {0}'.format(tmp_dir))
        run('psql -q -o /dev/null -d {0} < {1}/dump.sql'.format(db, tmp_dir))
        run('mv {0}/filestore /home/odoo/filestore/{1}'.format(tmp_dir, db))
        dbuuid = str(uuid4())
        run('''psql -tAq -d {0} -c "UPDATE ir_config_parameter SET value='{1}' WHERE key='{2}'"'''.format(db, dbuuid, 'database.uuid'))
        has_oauth = run('''psql -tAq -d {0} -c "SELECT EXISTS(SELECT 1 FROM pg_tables WHERE tablename = 'auth_oauth_provider')"'''.format(db))
        if has_oauth == 't':
            run('''psql -tAq -d {0} -c "UPDATE auth_oauth_provider SET client_id='{1}' WHERE id = (SELECT res_id FROM ir_model_data WHERE model='auth.oauth.provider' AND name='provider_openerp')"'''.format(db, dbuuid))
        create_date = run('''psql -tAq -d {0} -c "SELECT value FROM ir_config_parameter WHERE key='{1}'"'''.format(db, 'database.create_date')).strip()
        series = run('''psql -tAq -d {0} -c "SELECT replace((regexp_matches(latest_version, '^\d+\.0|^saas~\d+\.\d+|saas~\d+'))[1], '~', '-') FROM ir_module_module WHERE name='base'"'''.format(db))

    # create dns entries
    with settings(host_string='root@master.odoo.com'):
        run('''psql -tAq -d saas_master -c "INSERT INTO records(create_date, write_uid, name, content, type, ttl, prio, origin, domain_id) VALUES (now(), 1, '{host}', '{server}', 'MX', 600, 10, 'saas', 1), (now(), 1, '{host}', (SELECT public_ip FROM saas_server WHERE name='{server}'), 'A', 600, 0, 'saas', 1);"'''.format(host=host, server=server))

    # update metabase
    with settings(host_string='odoo@' + server):
        args = {
            'm': 'trial',
            's': 'activated',
            'p': get_port(series, with_sudo=False),
            'c': '"%s"' % create_date,
            'H': '"%s"' % host,
            'u': '"%s"' % dbuuid,
        }

        if trial_days:
            exp_date = (datetime.datetime.now() + relativedelta(days=int(trial_days), microsecond=0)).isoformat()
            args.update(m='trial', e='"%s"' % exp_date)

        args_str = ' '.join('-%s %s' % x for x in args.items())
        run('~/bin/oe-meta set -C {0} {1}'.format(args_str, db))

    params = werkzeug.urls.url_encode(dict(db=db, debug=1))
    echo(yellow('Restore complete, you can connect directly to install the saas_trial module and configure the oauth access:'))
    echo(green('https://{0}.odoo.com/login?{1}'.format(db, params)))
    echo(yellow("If this is a production database, don't forget to link it to a valid subscription!"))

@task
def extend_trial(db, days=7):
    env.forward_agent = True
    host = db + '.odoo.com'
    with settings(host_string='odoo@' + host):
        run('~/bin/oe-meta reset-cron-round {0}'.format(db))
        run("""psql -d meta -c "UPDATE databases
                                   SET mode='trial', status='',
                                       expire_date=expire_date + interval '%s days',
                                       cron_time=NULL
                                 WHERE name='%s'
                                   AND (mode='trial' OR (mode='drop' AND status='unsubscribed'))
                                   AND expire_date IS NOT NULL
                             RETURNING expire_date"
        """ % (days, db))

@task
def lock_db(db, *args, **kwargs):
    host = db + '.odoo.com'
    with settings(host_string='odoo@' + host):
        run("""oe-meta lock %s """ % db)
        print("\n")

@task
def unlock_db(db, *args, **kwargs):
    host = db + '.odoo.com'
    with settings(host_string='odoo@' + host):
        run("""oe-meta unlock %s """ % db)
        print("\n")

@task
def drop_db(db, force=False, only_new=False):
    """ Drop a db from its server """
    host = db + '.odoo.com'

    # Safety check
    if not force:
        answer = prompt(red("This will permanently drop %s. Are you sure?" % host),
                        key='drop_db',
                        validate=str2bool)
        if not answer:
            abort(yellow('You took the right decision'))

    # find db uuid
    with settings(host_string='odoo@' + host):
        meta = json.loads(run('oe-meta get -j {1}'.format(HOME, db)).strip())
        db_uuid = meta['uuid']
        db_status = meta['status']
        db_mode = meta['mode']
        db_create_date = datetime.datetime.strptime(meta['create_date'][:19], "%Y-%m-%dT%H:%M:%S")
        print(blue("Database {db} mode: [{db_mode}] status: [{db_status}]".format(
            db_mode=db_mode, db_status=db_status, db=db)))

    if only_new and db_create_date < datetime.datetime.now() - datetime.timedelta(weeks=2):
        abort(yellow("[new-only] Database is too old to be dropped: %s" % db_create_date))

    if db_status not in ('frozen',) and db_mode not in ('trial',):
        # remove database contract if any (click on "Disable") in our database
        with settings(host_string='odoo@ob.odoo.com'):

            if not force:
                is_contract_set = """
                    SELECT CASE
                             WHEN subscription_id IS NOT NULL THEN true
                             ELSE false
                           END AS has_contract
                    FROM openerp_enterprise_database
                    WHERE db_uuid='%s'
                """ % (db_uuid)
                has_contract = str2bool(run(""" psql -d openerp --pset=pager=off -Aq -t -c "%s" """ % is_contract_set).strip())
                if has_contract:
                    abort(yellow('You should not drop a db that has a subscription'))

            remove_oauth_users = """
                DELETE
                FROM oe_database_user_rel
                WHERE database_id = (SELECT id
                                     FROM openerp_enterprise_database
                                     WHERE db_uuid='%s')
                """ % (db_uuid)
            disable_database = """
                UPDATE openerp_enterprise_database
                SET subscription_id=NULL
                WHERE db_uuid='%s'
            """ % (db_uuid)
            run("""psql -d openerp -c "%s" """ % remove_oauth_users)
            run("""psql -d openerp -c "%s" """ % disable_database)

    # set db in drop mode
    with settings(host_string='odoo@' + host):
        drop_mode = 'frozen' if db_mode == 'frozen' else 'drop'
        run("""oe-meta lock %s """ % db)
        run("""oe-meta set -m %s -e '2016-01-01 00:00:00' %s """ % (drop_mode, db))
        #  # DISABLED, too slow, we can wait until next ping.
        #  run("""curl -L http://localhost:$(oe-meta last-port)/saas_worker/stats""")
        #  print(green("Database dropped!"))
        print(yellow("Database will be dropped within 5 minutes"))
        print("\n")

@task
def runstats():
    with settings(host_string='odoo@' + env.host_string):
        # update server stats
        run("""curl -L http://localhost:$(oe-meta last-port)/saas_worker/stats""")

@task
def set_free(db):
    """Sets a database back to free mode (extra_apps flag to False)"""
    host = db + '.odoo.com'
    with settings(host_string='odoo@' + host):
        run('~/bin/oe-meta set -A {0}'.format(db))

#----------------------------------------------------------
# misc
#----------------------------------------------------------

def ssh_public_key_to_fingerprints(line):
    # Returns all fingerprints for a public key
    # openssh 6.8 change the default representation of fingerprints:

    #   The default [FingerprintHash option] changes from MD5 to SHA256 and
    #   format from hex to base64.
    #   Fingerprints now have the hash algorithm prepended. An example of
    #   the new format: SHA256:mVPwvezndPv/ARoIadVY98vAC0g+P/5633yTC4d/wXE

    chunks = line.strip().split()
    for i, k in enumerate(chunks, 1):
        if k.startswith('ssh-'):
            key = chunks[i]
            break
    else:
        return []

    key = key.decode('base64')

    fp_md5 = hashlib.md5(key).hexdigest()

    return [
        # legacy md5 binary format (for servers with old openssh version)
        ':'.join(a + b for a, b in zip(fp_md5[::2], fp_md5[1::2])),
        # new formats
        'MD5:' + hashlib.md5(key).digest().encode('base64').strip('\n='),
        'SHA256:' + hashlib.sha256(key).digest().encode('base64').strip('\n='),
    ]

@task
@as_('root')
@with_settings(hide('stdout', 'running'))
def last_connections(limit=30):
    if limit:
        limit = int(limit)

    config = json.loads(run('cat /home/odoo/config/config.json'))
    slaves_ip = []
    me = config['me']['name']
    slaves_ip = [
        w['private_ip'] or w['public_ip']
        for w in config['pong']['workers'].values()
        if w['master'] == me
    ]

    fp2user = {}
    with hide('stderr'):
        keys = '\n'.join([
            run("cat ~/.ssh/authorized_keys", warn_only=True),
            run("cat ~/.ssh/authorized_keys2", warn_only=True),
            run("cat ~/.ssh/authorized_keys3", warn_only=True),
            run("cat %s/.ssh/authorized_keys3" % HOME, warn_only=True),
        ])

    for k in keys.splitlines():
        k = k.strip()
        if not k or k.startswith('#'):
            continue
        user = k.split()[-1]
        fp2user.update(dict.fromkeys(ssh_public_key_to_fingerprints(k), user))

    fingerprints = run(
        'grep -F --no-filename "Accepted publickey for" /var/log/auth.log.1 /var/log/auth.log'
        ' | grep -Ev " from (%s) port "'
        ' | tail -%s' %
        ('|'.join(slaves_ip), limit)
    ).splitlines()

    for fpl in fingerprints:
        fpl = fpl.split()
        fp = fpl[-1]
        date = ' '.join(fpl[:3])
        user = fpl[8]
        puts("%s %s as %s - %s" % (date, fp2user.get(fp, "?"), user, fpl))

@task
@hosts('root@master.odoo.com')
@with_settings(hide('stdout', 'running'))
def get_roles():
    """generate role definition file"""

    def run_query(query):
        return run('psql -X --pset=pager=off -tAF, -d saas_master -c "%s"' % query).strip().split('\n')

    wanted_roles = 'trial demo bak slave lb of ob runbot'.split()
    sql_roles = 'ARRAY%r' % (wanted_roles,)
    query = "SELECT name, mode FROM saas_server WHERE %s && regexp_split_to_array(mode, ' ') ORDER BY name;" % sql_roles

    roles = defaultdict(list)
    for m in run_query(query):
        host, modes = m.split(',', 1)
        for mode in modes.split():
            if mode in wanted_roles:
                roles[mode].append('root@%s' % host)

    roles['slave'] = [h for h in roles['slave'] if h in roles['trial']]      # only trial slaves
    roles['trial-only'] = roles['trial'] = [h for h in roles['trial'] if h not in roles['slave']]

    query = "SELECT 'root@' || name FROM saas_server WHERE lastping IS NOT NULL ORDER BY name;"
    roles['all'] = list(map(str.strip, run_query(query)))

    env.roledefs = dict(roles)

    with open(roledefs_file, 'w') as f:
        json.dump(dict(roles), f)

@task
def list_roles(list_servers='f'):
    list_servers = str2bool(list_servers)
    for role, servers in env.roledefs.items():
        s = blue(role)
        if list_servers:
            s += ': ' + ', '.join(x.partition('@')[2].replace('.odoo.com', '') for x in servers)
        puts(s, show_prefix=False)

@task
def fetch_db(db, todir='/tmp/', live=False):
    live = str2bool(str(live))
    host = db + '.odoo.com'
    with settings(host_string='odoo@%s' % host):
        path = '{0}/bakdb/{1}.*.sql.gz'.format(HOME, db)
        if live:
            path = '/tmp/{0}.sql.gz'.format(db)
            run('pg_dump -O -Fp {0} | gzip > {1}'.format(db, path))
        with settings(hide('stdout', 'running')):
            sz = int(run('stat --printf "%s" {0} 2>/dev/null || echo -n 0'.format(path)))
        puts(green("Backup Size: ") + sizeof_fmt(sz), show_prefix=False)
        get(path, os.path.join(todir, "{}.sql.gz".format(db)))
        if live:
            run('rm {0}'.format(path))

@task
@with_settings(hide('stdout', 'running'))
@hosts('root@ob.odoo.com')
def gdprinfo(email, name=None):
    import prettytable
    echo = partial(puts, show_prefix=False)

    if name is None:
        name = email

    if '@' not in email and '@' in name:
        name, email = email, name

    def query(q):
        return run("""psql -X --pset=pager=off -Aq -H -c "{0}" -d openerp""".format(q))

    def pretty(result):
        echo(prettytable.from_html_one(result))

    q = """
        select 'lead', id, email_from, active, 'name: ' || name from crm_lead where email_from ilike '%{email}%' or name ilike '{name}' UNION ALL
        select 'partner', id, email, active, 'name: ' || name from res_partner where email ilike '%{email}%' or name ilike '{name}' UNION ALL
        select 'user', id, login, active, 'partner_id: ' || partner_id::text from res_users where
                (
                 (login ilike '%{email}%') OR
                 (partner_id in (select id from res_partner where email ilike '%{email}%' or name ilike '{name}'))
                )
        UNION ALL
        select 'applicant', id, email_from, active, 'job_id: ' || job_id::text   from hr_applicant where email_from ilike '%{email}%' or name ilike '{name}' UNION ALL
        select 'attendee', id, email, true, 'event_id: ' || event_id::text  from event_registration where email ilike '%{email}%' or name ilike '{name}' UNION ALL
        select 'mass contact', id, email, true, 'name: ' || name  from mailing_contact where email ilike '%{email}%' or name ilike '{name}' UNION ALL
        select 'challenge', id, login, active, 'login: ' || login  from challenge_user where login ilike '%{email}%' UNION ALL
        select 'task', id, email_from, active, 'name: ' || name  from project_task where email_from ilike '%{email}%' UNION ALL
        select 'db', id, company_email, true, 'url: ' || url || ' last ping: ' || last_ping from openerp_enterprise_database where (company_email ilike '%{email}%' or company_name ilike '%{name}%') UNION ALL
        select 'db user', id, login, true, 'name: ' || name from openerp_enterprise_database_user where (login ilike '%{email}%' or name ilike '%{name}%')
    """.format(name=name.strip(), email=email.strip())
    r1 = query(q)

    q = """
        select 'task', id, name from project_task where email_from ilike '%{email}%' or partner_id in
                (select id from res_partner where email ilike '%{email}%' or name ilike '{name}')  UNION ALL
        select 'invoice', id, name from account_move
                where (
                 partner_id in (select id from res_partner where email ilike '%{email}%' or name ilike '{name}')
                 OR
                 partner_shipping_id in (select id from res_partner where email ilike '%{email}%' or name ilike '{name}')
                 OR
                 commercial_partner_id in (select id from res_partner where email ilike '%{email}%' or name ilike '{name}')
                )
                UNION ALL
        select 'Subscription', id, name from sale_subscription where partner_id in
                (select id from res_partner where email ilike '%{email}%' or name ilike '{name}')  UNION ALL
        select 'SO', id, name from sale_order where partner_id in
                (select id from res_partner where email ilike '%{email}%' or name ilike '{name}')  UNION ALL
        select 'PO', id, name from purchase_order where partner_id in
                (select id from res_partner where email ilike '%{email}%' or name ilike '{name}')  UNION ALL
        select 'task', id, name from project_task where partner_id in
                (select id from res_partner where email ilike '%{email}%' or name ilike '{name}')  UNION ALL
        select 'related db user', id, login from openerp_enterprise_database_user where user_id in
                (
                    select id from res_users where
                    (
                     (login ilike '%{email}%') OR
                     (partner_id in (select id from res_partner where email ilike '%{email}%' or name ilike '{name}'))
                    )
                ) UNION ALL
        SELECT 'message', id, model || ' -> ' || res_id::text from mail_message where author_id in
                (select id from res_partner where email ilike '%{email}%' or name ilike '{name}')

    """.format(name=name.strip(), email=email.strip())
    r2 = query(q)

    echo(green('Basic Personal Data Records'))
    pretty(r1)
    echo(green('Indirect Personal Data References'))
    pretty(r2)

@task
@with_settings(hide('stdout', 'running'))
@hosts('root@master.odoo.com')
def serverinfo(host):
    if "." not in host:
        host += ".odoo.com"
    ip = gethostbyname(host)
    q = """SELECT name, public_ip, mode, stat_revs,
                  (SELECT string_agg(name, E'\n') as backups
                     FROM saas_server
                     WHERE backup_group_id = s1.backup_group_id
                           AND mode like '%bak%')
           FROM saas_server s1
          WHERE public_ip = '{0}'""".format(ip)
    output = run("""psql -X --pset=pager=off -AqH -c "{0}" -d {1}""".format(q, 'saas_master'))
    import prettytable
    puts(prettytable.from_html_one(output), show_prefix=False)

@task
@with_settings(hide('stdout', 'running'))
@hide_disconnect_messages
def dbinfo(db):
    if db.endswith('.odoo.com'):
        db = db[:-9]
    assert(db)

    def query(q, db=db, html=False):
        extra_args = '-H' if html else '-t'
        return run("""psql -X --pset=pager=off -Aq {2} -c "{0}" -d {1}""".format(q, db, extra_args))

    echo = partial(puts, show_prefix=False)

    with cd('/tmp'), settings(host_string='root@%s.odoo.com' % db):
        hostname = run('hostname')

        dbuuid = query("select value from ir_config_parameter where key='database.uuid'")
        base_version = query("select latest_version from ir_module_module where name='base'")

        meta = json.loads(sudo('{0}/bin/oe-meta get -j {1}'.format(HOME, db)).strip())

        with settings(hide('warnings'), warn_only=True):
            sz = int(run('stat --printf "%s" {0}/bakdb/{1}.sql.gz 2>/dev/null || echo -n 0'.format(HOME, db)))
            fsz = run('du -sh /home/odoo/filestore/{0}/'.format(db)).split()[0]
            fields = "id,oauth_uid,login,password,active,share,create_date,write_date,login_date"
            user_html = query("SELECT {0} FROM res_users ORDER BY id LIMIT 10".format(fields), html=True)
        if user_html.failed:
            with settings(hide('warnings'), warn_only=True):
                # share column not present. try without
                fields = "id,oauth_uid,login,password,active,create_date,write_date,login_date"
                user_html = query("SELECT {0} FROM res_users ORDER BY id LIMIT 10".format(fields), html=True)
            if user_html.failed:
                # login_date column maybe not present. try without (gone in 9.0)
                fields = "id,oauth_uid,login,password,active,create_date,write_date"
                user_html = query("SELECT {0} FROM res_users ORDER BY id LIMIT 10".format(fields), html=True)

    # output
    echo(green("Hostname: ") + hostname)
    echo(green("Database UUID: ") + dbuuid)
    echo(green("Base Version: ") + base_version)
    echo(green("Last Backup Size: ") + (sizeof_fmt(sz) if sz else 'N/A'))
    echo(green("Filestore Size: ") + fsz)
    echo(green("Meta"))
    for k in "port mode status create_date expire_date last_cnx_date extra_apps \
              cron_round cron_time email_daily_limit email_daily_count email_total_count".split():
        value = meta[k]
        if value is None:
            value = ""
        if k.endswith(('_date', '_time')):
            value = value.replace('T', ' ')
        echo("\t{0}: {1}".format(blue(k), value))
    echo("\t{0}:".format(blue('Hosts')))
    for h in meta['hosts']:
        echo('\t\t' + h)

    import prettytable
    pt = prettytable.from_html_one(user_html)
    echo(green('Users'))
    echo(pt)
    row = pt._get_rows(pt._get_options(dict(end=1)))[0]
    admin_login, admin_password = row[2], row[3]

    if admin_password:
        params = werkzeug.urls.url_encode(dict(db=db, login=admin_login, key=admin_password, debug=1))
        echo(green("Direct admin login: ") + "https://{0}.odoo.com/login?{1}".format(db, params))
    else:
        echo(yellow("No local admin password - direct connection not possible"))

    with settings(host_string='odoo@ob.odoo.com'):
        oauth_logins = query("""SELECT u.id as oauth_uid, u.login as email
                                FROM oe_database_user_rel o
                                    JOIN openerp_enterprise_database d ON d.id = o.database_id
                                    JOIN res_users u ON u.id = o.user_id
                                WHERE d.db_uuid = '%s' and o.is_admin""" % dbuuid, db="openerp", html=True)
        if oauth_logins:
            echo(green("Admin OAuth Login(s): "))
            echo(prettytable.from_html_one(oauth_logins))
        else:
            oauth_login = None
            echo(yellow("No Admin OAuth Login. Is it a trial database?"))

        q = "SELECT d.subscription_id, s.client_order_ref, p.id, p.email \
               FROM openerp_enterprise_database d, sale_order s, res_partner p \
              WHERE db_uuid = '{0}' \
                AND d.subscription_id = s.id \
                AND s.partner_id = p.id".format(dbuuid)

        contracts = query(q, db='openerp')
        if contracts:
            contract_id, contract, customer_id, customer_email = contracts.split('|')
            echo(green("Contract: ") + "#%s, %s" % (contract_id, contract))
            echo(green("Customer: ") + "#%s, %s" % (customer_id, customer_email))
        else:
            contract = None
            echo(yellow("No contract for this database"))

        # error checks
        mode = meta['mode']
        if contract and mode != 'paid':
            echo(red('ERROR: The database has a contract but is not in paid mode in meta base'))
        if not contract and mode == 'paid':
            echo(red('ERROR: The database is in paid mode in meta base but has no contract'))
        if contract and oauth_login != customer_email:
            echo(red('ERROR: The Administrator of the database is not the customer of the contract'))


@task
def cron_rounds():
    run('psql -X --pset=pager=off -d meta -c "SELECT * FROM cron_rounds"')

@task
@as_('root')
def dblost(dropthem=False):
    # XXX keep query in sync with munin plugin openerp_dbcount
    query = """SELECT d.datname FROM pg_database d
            LEFT JOIN databases m ON d.datname = m.name
                WHERE d.datistemplate = false
                  AND m.name IS NULL
                  AND pg_get_userbyid(d.datdba) = 'root'
                  AND d.datname !~ '^(_?demo|_migrate|_failed_migration|_migrated)_.+'
                  AND d.datname != 'meta'
            """
    run('psql -X --pset=pager=off -d meta -c "%s"' % query)
    if dropthem:
        run('psql -X --pset=pager=off -tA -d meta -c "%s" | xargs -rtn1 dropdb' % query)

@task
@runs_once
@roles('trial')
@hide_disconnect_messages
def count_dbs():
    """ # of databases per running branch """

    trial_servers = ['root@%s' % t.split('@')[-1] for t in env.hosts]
    if not trial_servers:
        trial_servers = env.roledefs['trial']
    result = defaultdict(int)
    b2s = defaultdict(set)
    t = len(trial_servers)
    for i, trial in enumerate(trial_servers, 1):
        name = trial.split('@')[1].split('.')[0]
        puts('processing %s [%d/%d]' % (name, i, t), show_prefix=False)
        with settings(host_string=trial):
            with hide('output', 'running'):
                branches = all_branches()
                if not branches:
                    continue

                branches_ports = dict.fromkeys(branches, 0)
                for branch in branches:
                    branches_ports[branch] = get_port(branch)

            CASE_BRANCH_PORT = ' '.join(
                ["WHEN %s THEN '%s'" % (p, b) for b, p in branches_ports.items()]
            )

            COUNT_DBS_BY_PORT = """
                SELECT CASE port """ + CASE_BRANCH_PORT + """
                       ELSE 'not running' END AS version,
                       COUNT(port) AS total
                  FROM databases
                 WHERE mode in ('trial', 'paid')
              GROUP BY port
              ORDER BY port;
            """

            with hide('running', 'output'):
                versions = run('psql -X -tAF, --pset=pager=off -d meta -c "%s"' % COUNT_DBS_BY_PORT)
                for v in versions.strip().splitlines():
                    p, c = v.strip().split(',')
                    result[p] += int(c)
                    b2s[p].add(name)

    puts(green('=' * 79), show_prefix=False)
    for v, c in result.items():
        s = ','.join(b2s[v])
        puts('%s: %s {%s}' % (yellow(v), c, s), show_prefix=False)

@task
@as_('root')
def count_dbs_per_mode(branch):
    with hide('output', 'running'):
        if branch not in all_branches():
            return
        port = get_port(branch)
    with hide('running'):
        query = "SELECT mode, count(*) FROM databases WHERE port=%s GROUP BY 1 ORDER BY 1" % port
        run('psql -X --pset=pager=off -d meta -c "%s"' % query)


@task
def get_custom_domains():
    """Prints tcustom domains registered on the given trial servers.

    Outputs a list prefixed with '##' to make it easy to grep. For example extracting
    finding all customers using Cloudflare NS could be done like this:

      $ fab -Pz 10 -w  -R trial-only get_custom_domains > /tmp/custom
      $ grep -oP "(?<=## )(.*)" /tmp/custom | xargs -n1 host -t NS > /tmp/custom_ns
      $ grep cloudflare /tmp/custom_ns | cut -d' ' -f 1 | sort | uniq
    """
    with settings(host_string='odoo@' + env.host):
        domains = run(
            """psql -X --pset=pager=off -At -c """
            """ "SELECT domains.name || ' (' || databases.name || ') ' """
            """    FROM domains """
            """    JOIN databases ON (domains.db_id = databases.id) """
            """   WHERE databases.mode !='hostname' ;" -d meta""").splitlines()
    # Use the '##' patten to grep actual entries for postprocessing...
    puts("\n## %s\n" % "\n## ".join(domains))

@task
@runs_once
@roles('trial')
def count_installed_modules(where="mode in ('paid')"):
    """
    fabo count_installed_modules:"mode in ('paid') and create_date >\= '2014-01-01'"
    """
    trial_servers = [t.split('@')[-1] for t in env.hosts]
    result = {}
    for trial in trial_servers:
        with hide('running', 'output'):
            with settings(host_string='odoo@' + trial):
                puts('Now on %s' % trial)
                databases = run('psql -tA -d meta -c "SELECT array_agg(name) FROM databases %s"' % ('WHERE %s' % where) if where else '').strip()[1:-1].split(',')
                databases_count = len(databases)
                puts('%s databases' % databases_count)
                i = 0
                for database in databases:
                    i += 1
                    if i % 10 == 0:
                        puts('%.2f %%' % (float(i) / databases_count * 100))
                    database = database.strip()
                    installed_modules = run('psql -tA -d %s -c "SELECT array_agg(name) FROM ir_module_module WHERE state = \'installed\'"' % (database)).strip()[1:-1].split(',')
                    for module in installed_modules:
                        module = module.strip()
                        result[module] = result.get(module, 0) + 1

    for module_count in sorted(result.items(), key=lambda x: x[1], reverse=True):
        puts('%s,%s' % (module_count[0], module_count[1]), show_prefix=False)


@task
@runs_once
@roles('trial')
def list_deserted_dbs(where="mode in ('paid')"):
    """
    fabo count_installed_modules:"mode in ('paid') and create_date >\= '2014-01-01'"
    """
    trial_servers = [t.split('@')[-1] for t in env.hosts]
    result = {}
    for trial in trial_servers:
        with hide('running', 'output'):
            with settings(host_string='odoo@' + trial):
                puts('Now on %s' % trial)
                databases = filter(lambda x: x and x != '""' and x != 'NULL', run('psql -tA -d meta -c "SELECT array_agg(name||\'#\'||uuid) FROM databases %s"' % ('WHERE %s' % where) if where else '').strip()[1:-1].split(','))
                databases = [database.split('#') for database in databases]
                databases_count = len(databases)
                puts('%s databases' % databases_count)
                i = 0
                for database in databases:
                    db_name = database[0]
                    i += 1
                    if i % 10 == 0:
                        puts('%.2f %%' % (float(i) / databases_count * 100))
                    db_name = db_name.strip()
                    login_date = run('psql -tA -d %s -c "SELECT max(login_date) FROM res_users"' % (db_name)).strip()
                    if not login_date or not datetime.datetime.strptime(login_date, '%Y-%m-%d') >= datetime.datetime.now() - relativedelta(days=75):
                        result.setdefault(trial, []).append((database, login_date))
    trials = sorted(result.items())
    uuids = []
    for trial in trials:
        puts('---------------\n' + trial[0] + '\n---------------', show_prefix=False)
        for db in sorted(trial[1], key=lambda x: x[1]):
            puts('%s: %s' % (db[0], db[1]), show_prefix=False)
            uuids.append("'%s'" % db[0][1])
    puts('-----------------', show_prefix=False)
    for trial in trials:
        puts('%s: %s' % (trial[0], [db[0][1] for db in trial[1]]), show_prefix=False)

    with settings(host_string='odoo@ob.odoo.com'):
        contracts = [x for x in run('psql -tA -d openerp -c "SELECT array_agg(c.code) FROM openerp_enterprise_database d, account_analytic_account c WHERE d.analytic_account_id = c.id AND d.db_uuid in (%s)"' % ",".join(uuids)).strip()[1:-1].split(',') if x and x != '""']
        puts(contracts, show_prefix=False)

@task
def bootstrap_pg_replication():
    for version in (9.1, 9.2, 9.3):
        datadir = '/home/postgres/%s/main' % version
        etcdir = '/etc/postgresql/%s/main' % version
        if not fabtools.files.is_dir(datadir) or not fabtools.files.is_dir(etcdir):
            continue
        if fabtools.files.is_file('{0}/key.pem'.format(datadir)) or fabtools.files.is_file('{0}/cert.pem'.format(datadir)):
            continue
        sudo('openssl req -x509 -nodes -newkey rsa:2048 -keyout {0}/key.pem -out {0}/cert.pem -days 3650 -subj "/CN={1}"'.format(
            datadir, fabtools.system.get_hostname()), user='postgres')
        sudo('chmod 600 {0}/key.pem'.format(datadir), user='postgres')
        ip = [fabtools.network.address(i) for i in fabtools.network.interfaces()][0]
        sudo('echo "listen_addresses = \'localhost, {0}\'" >> {1}/postgresql.local.conf'.format(
            ip, etcdir), user='postgres')

@task
@hosts('root@master.odoo.com')
def create_custom_dns(db, domain):
    assert '.' not in db
    base_cmd = 'dig {0} {1} +short'
    ttl = 600

    def gen_query_values(name, type, content, ttl, prio):
        return "('{name}', '{type}', '{content}', {ttl}, {prio}, 'saas', 1)".format(**locals())

    ip = gethostbyname(db + '.odoo.com')

    with cd('/tmp'), settings(host_string='root@%s.odoo.com' % db):
        hostname = run('hostname')

    # SOA, A, www CNAME
    base_values = ','.join(
        [gen_query_values(n, t, c, ttl, 0) for n, t, c in
            [(domain, 'SOA', 'ns1.odoo.com admin@odoo.com 0'), (domain, 'A', ip), ('www.' + domain, 'CNAME', domain)]])
    # existing MX records
    mx_values = ','.join(
        [gen_query_values(domain, 'MX', c, ttl, p) for p, c in
            [tuple(r.split()) for r in local(base_cmd.format('MX', domain), capture=True).splitlines()]])
    # fallback MX record
    if not mx_values:
        mx_values = gen_query_values(domain, 'MX', hostname, ttl, 10)

    query = 'INSERT INTO records(name, type, content, ttl, prio, origin, domain_id) VALUES {0}, {1};'.format(base_values, mx_values)
    run('psql -X --pset=pager=off -tAF, -d saas_master -c "%s"' % query).strip().split('\n')


# ======
@task
@roles('runbot')
def runbot_kill_bad_crons():
    # when installing a module from apps, the server is restarted without "-d" flag, which lead the
    # cron worker to process all databases, which generate intrusive warning messages in runbot log
    run("grep -aE 'runbot\/static\/build\/.* --xmlrpc-port=' /proc/*/cmdline | grep -av gevent | grep -vaP ' -d ' | cut -d/ -f3 | xargs -rtn1 kill")

@task
@roles('runbot')
def update_runbot():
    path = os.path.join(HOME, 'runbot/extra')
    # run("chown -R {0}:{0} {1}".format(USER, path))

    with cd(path):
        sudo('git fetch --quiet --prune')
        sudo('git rebase --quiet --autostash')
        with hide('status', 'commands'):
            no_conflicts = sudo('git diff --diff-filter=U --no-patch --exit-code').succeeded

    if no_conflicts:
        run('systemctl restart runbot')
    else:
        abort(red('Conflicts occurs in runbot repo'))

# =========================================
# git helpers methods (use github api)
import requests

GH_TOKEN = "e44cc91ee06bcd9adc153af205653a1cf0e26f6f"
GH_Session = requests.Session()
GH_Session.auth = (GH_TOKEN, 'x-oauth-basic')
GH_Session.headers.update({'Accept': 'application/vnd.github.v3+json'})

def github(method, path, **kw):
    endpoint = 'https://api.github.com'
    return GH_Session.request(method, endpoint + path, **kw)

@task
def gh_terminate_user(*users):
    for user in users:
        for org in ['odoo', 'odoo-dev']:
            puts('%s: %s' % (org, github('DELETE', '/orgs/{}/memberships/{}'.format(org, user)).text))


@task
def count_records(table, branch):
    """ count_records:table,branch """

    echo = partial(puts, show_prefix=False)
    spin = r'/-\|'

    db_rec = {}
    with hide('running', 'output'):
        dbs = sudo('~/bin/oe-meta list -m paid {0}'.format(branch), user='odoo').splitlines()
        echo(green('db total count: {0}'.format(len(dbs))))
        if not dbs:
            return
        for i, db in enumerate(dbs, 1):
            puts('\r%s %d/%d' % (spin[i % len(spin)], i, len(dbs)), show_prefix=False, end='', flush=True)
            res = run("""psql -tAq -d {0} -c "SELECT count(1) FROM information_schema.tables WHERE table_name = '{1}' AND table_type = 'BASE TABLE'" """.format(db, table))
            if (int(res) > 0):
                db_rec[db] = int(run("""psql -tAq -d {0} -c "SELECT count(id) FROM {1}" """.format(db, table)))
    echo('')

    non_zero = 0
    ten_plus = 0
    for db, count in db_rec.items():
        if count > 0:
            echo(yellow('db : {0}, records: {1}'.format(db, count)))
            non_zero += 1
            if count >= 10:
                ten_plus += 1
    echo(green('dbs with records in {0}: {1} ({2} with 0 record)'.format(table, non_zero, len(db_rec.keys()))))
    pc = 100 * non_zero / (1.0 * len(dbs))
    pc_10 = 100 * ten_plus / (1.0 * len(dbs))
    echo(green('dbs with > 0 records: {0:.2f}%; dbs with > 10 records: {1:.2f}%'.format(pc, pc_10)))
    echo(green('records sum: {0}'.format(sum(db_rec.values()))))


@task
def report_databases():
    """Report the number of databases on servers per version"""
    table = ['Version Paid Trial Drop Total'.split()]
    with settings(hide('stdout', 'running', 'warnings')):
        hd = run("df -h | grep -e '/home$' || df -h | grep -e '/$'").split()
        for version in all_branches():
            line = [version]
            for mode in ('paid', 'trial', 'drop'):
                line.append(int(sudo('oe-meta count -m %s %s' % (mode, version))))
            line.append(sum(line[1:]))
            table.append(line)

    grand_total = sum(row[-1] for row in table[1:])
    title = ("%s (Total databases: %d) (HD space: %s/%s[%s] - %s free)"
             % (env.host, grand_total, hd[2], hd[1], hd[4], hd[3]))
    import tabulate  # noqa
    stable = tabulate.tabulate(table, headers='firstrow', tablefmt='pipe')
    print('\n%s\n%s\n\n%s\n\n' % (title, ('=' * len(title)), stable))

# --------------------------------------------------------------------------- .fabroles management
roledefs_file = os.path.join(HERE, '.fabroles')
if not os.path.isfile(roledefs_file):
    warn(yellow('role definition file does not exists. Executing `get_roles` task'))
    execute(get_roles)
else:
    MAX_DAYS = 1500
    if os.path.getmtime(roledefs_file) < time.time() - (MAX_DAYS * 86400):
        warn(yellow('role definition file is older than %d days. Executing `get_roles` task' %
                    MAX_DAYS))
        execute(get_roles)
    else:
        with open(roledefs_file, 'r') as r:
            env.roledefs = json.load(r)


if __name__ != '__main__' and len(sys.argv) == 1:
    os.system('%s -l' % sys.argv[0])
    sys.exit(0)

# vim:et:
